grammar SortVerilog;

    
options {
  backtrack = true;
  k         = 2;
  output    = template;
  rewrite   = true;
}

@header {
  	package com.eu.miscedautils.SortHDL;
  	import com.eu.miscedautils.CommonUtils.LogMessages; 
//  	import  com.eu.miscedautils.rtlmodifier.ModuleMap;
//  	import  com.eu.miscedautils.rtlmodifier.UniquifyRTL;
//  	import  com.eu.miscedautils.rtlmodifier.ExcludeModules; 
  	import java.util.Iterator;
  	import java.util.List;
  	import java.util.ArrayList;
  	import java.util.LinkedList;
  	import org.antlr.runtime.ANTLRFileStream;
	import org.antlr.runtime.ANTLRStringStream;
	import org.antlr.runtime.CharStream;
	import org.antlr.runtime.CommonTokenStream;
	import org.antlr.runtime.RecognitionException;
	import org.antlr.runtime.TokenSource;
	import org.antlr.runtime.TokenStream;
	import java.io.File;
	import java.io.FileInputStream;
	import java.io.IOException;
}

@lexer::header {
  	package com.eu.miscedautils.SortHDL;
}

@lexer::members {
	boolean isSequentialBlock = false;
	boolean isAssign = false;
	boolean isCompilerDirective = false;
}

@members 
{	
  private SortHDLUnit currMod;
	@Override
	public void emitErrorMessage(String msg) 
	{
	}

	@Override
	public String getErrorMessage(RecognitionException e, String[] tokenNames)
	{
		List stack = getRuleInvocationStack(e, this.getClass().getName());
		String msg = null;
		if ( e instanceof NoViableAltException ) 
		{
			NoViableAltException nvae = (NoViableAltException)e;
			msg = " no viable alt; token="+e.token+
			" (decision="+nvae.decisionNumber+
			" state "+nvae.stateNumber+")"+
			" decision=<<"+nvae.grammarDecisionDescription+">>";
		}
		else 
		{
			msg = super.getErrorMessage(e, tokenNames);
		}
		return stack+" " + msg;
	}
	
	public String getTokenErrorDisplay(Token t) 
	{
		return t.toString();
	}
}
 
  
parseVerilog 
:
  (description)* EOF
  ;

description
  :
  module
  | udp
  | directive
  ;

module 
  :
  (
    MODULE
    | MACROMODULE
  )
  nm=name_of_module 
  {
        SortHDLRoot.instance().setLastVlogModule( null );
        if ( null != $nm.tok )
        {
          SortHDLUnit du = SortHDLRoot.instance().getDesignUnit($nm.tok.getText());
          HDLFile file = SortHDLRoot.instance().getHDLFile();
          if ( null == du )
          {
        	  du = new SortHDLUnit(SortHDLBase.ObjType.MODULE, $nm.tok.getText(), file, $nm.tok.getLine());
            SortHDLRoot.instance().addDesignUnit( du );             
          }
          else 
          {
            if ( SortHDLBase.ObjType.UNDEFINED_MODULE == du.getObjType() )
            {
              du.setObjType( SortHDLBase.ObjType.MODULE );
              du.setFile( file );
              du.setLine( $nm.tok.getLine() );
            }
            else
            {
              LogMessages.instance().print(5085,
                                            LogMessages.MSG_SEVERITY.ERROR, 
                                            "Design unit '" + $nm.tok.getText() + "'is multiply defined ...");  
            } 
          }
          SortHDLRoot.instance().setLastVlogModule( du );           
        } 
  } 
(  
  (
    ( 
        list_of_ports
    )? SEMI
  )  
  |
  (
    (
    POUND LPAREN  
    (
        p1=parameter_declaration_no_semi
        |
        lp1=localparam_declaration_no_semi
    )
    (   COMMA
        p2=parameter_declaration_no_semi
        |
        lp2=localparam_declaration_no_semi
    )*
    RPAREN
    )?
    (
        ( LPAREN  
        ( port_declaration_no_semi
        )
        ( COMMA
            port_declaration_no_semi
        )*
        RPAREN
        )
        |
        list_of_ports
    )
    ? SEMI  
  )
)
  (module_item)* 
  ENDMODULE     
  ;

port_declaration_no_semi
    :
    input_declaration_no_semi
    | 
    output_declaration_no_semi 
    | 
    inout_declaration_no_semi
    ;
    

list_of_ports
  :
  LPAREN p1=port (COMMA p2=port)* RPAREN
  ;

port
  :
  (port_expression)?
  | DOT name_of_port LPAREN (port_expression)? RPAREN
  ;

port_expression
  :
  port_reference
  | LCURLY p1=port_reference (COMMA p2=port_reference)* RCURLY
  ;

port_reference
  :
  (name_of_variable LBRACK expression COLON) => name_of_variable LBRACK expression COLON expression RBRACK
  | (name_of_variable LBRACK) => name_of_variable LBRACK expression RBRACK
  | name_of_variable
  ;

module_item 
  :
  // ambiguity between net_declaration and continuous_assign,
  // but parser gets it right: keyword chosen over IDENTIFIER.
  p1=parameter_declaration
  | input_declaration
  | output_declaration
  | inout_declaration
  | net_declaration
  | // Already covered
   reg_declaration
  | // Already covered
   time_declaration
  | integer_declaration
  | real_declaration
  | event_declaration
  | gate_declaration
  | instantiation 
  | parameter_override
  | continuous_assign
  | // Already covered
   specify_block
  | initial_statement
  | always_statement
  | task
  | function
  | directive
  | generate_stmt 
  ;

generate_stmt
  :
  GENVAR IDENTIFIER SEMI GENERATE genarate_block (BEGIN)? (COLON name_of_block (block_declaration)*)? (generate_item)* (END )? ENDGENERATE
  ;

genarate_block
  :
  (FOR LPAREN assignment SEMI expression SEMI assignment RPAREN)
  | ( IF LPAREN expression RPAREN)
  ;

generate_item 
  :
  continuous_assign
  | parameter_declaration net_declaration
  | // Already covered
   reg_declaration
  | // Already covered
   time_declaration
  | integer_declaration
  | real_declaration
  | event_declaration
  | gate_declaration
  | instantiation 
  | parameter_override
  | specify_block
  | initial_statement
  | always_statement
  | task
  | function
  | directive
  ;

instantiation 
  :
  (module_instantiation) => module_instantiation
 // |udp_instantiation
  ;
  
//udp_instantiation :
//  name_of_UDP (drive_strength)? (delay)?
//          udp_instance ( COMMA udp_instance )* SEMI
//  ;

//----------------------------------------------------------------------------
// UDP specs
//----------------------------------------------------------------------------

udp
  :
  PRIMITIVE name_of_UDP LPAREN name_of_variable (COMMA name_of_variable)* RPAREN SEMI (udp_declaration)+ (udp_initial_statement)? table_definition ENDPRIMITIVE
  ;

udp_port_list
  :
  udp_name_of_port (COMMA udp_name_of_port)*
  ;

udp_declaration
  :
  output_declaration
  | input_declaration
  | reg_declaration
  ;

udp_initial_statement
  :
  // Baya        'initial' output_terminal_name ASSIGN init_val SEMI
  INITIAL output_terminal_name ASSIGN NUMBER SEMI
  ;

// Use a semantic predicate to determine whether a matched NUMBER
// is a valid special value in the given context.
// This kludge avoids having the special values in the Literals table,
// thus avoiding a lexical conflict.

//Baya : Below is commented by Baya otherwise we get error on
// continuous assignment like assign lhs = 1'b1 or assign lhs = 1'b0

//init_val :
//        '1\'b0' |
//        '1\'b1' |
//        '1\'bx' |
//	n=NUMBER
//	{ $n.text=="0" || $n.text=="1"}?
//	;

table_definition
  :
  TABLE table_entries ENDTABLE
  ;

// Don't try to parse table entries; just collect them.
// There are ambiguities between edge_symbol and level_symbol,
// and textbook Verilog examples don't seem to follow rules
// completely. For example,
//	'0    00    :    0;'
// doesn't match grammar because of '00', but is frequently used.

table_entries
  :
  //(sequential_entry) => (sequential_entry)+ |
  //(combinational_entry)+
  (
    (
      ~(
        SEMI
        | ENDTABLE
       )
    )+
    SEMI
  )*
  ;

task
  :
  TASK name_of_task SEMI (tf_declaration)* statement_or_null ENDTASK
  ;

function
  :
  FUNCTION (range_or_type)? name_of_function SEMI (tf_declaration)+ statement ENDFUNCTION
  ;

range_or_type
  :
  range
  | INTEGER
  | REAL
  ;

tf_declaration
  :
  parameter_declaration
  | output_declaration
  | input_declaration
  | inout_declaration
  | reg_declaration
  | time_declaration
  | integer_declaration
  | real_declaration
  | event_declaration
  ;

//----------------------------------------------------------------------------
// Declarations
//----------------------------------------------------------------------------
parameter_declaration : parameter_declaration_no_semi SEMI
    ;

parameter_declaration_no_semi : 
       l1=PARAMETER 
       ( INTEGER 
       |
       'signed'
       |
       REAL 
       )?
       (r1=range)? 
       la1=list_of_param_assignments
       ;

localparam_declaration : localparam_declaration_no_semi SEMI
    ;

localparam_declaration_no_semi :
    l1=LOCALPARAM (r1=range)? list_of_localparam_assignments
    ;

list_of_param_assignments :
    p1=param_assignment
    ( COMMA 
        p2=param_assignment 
    )*
    ;
    
list_of_localparam_assignments :
    p1=localparam_assignment
    ( COMMA 
        p2=localparam_assignment 
    )*
    ;

param_assignment : identifier ASSIGN expression
   ;

localparam_assignment : identifier ASSIGN expression
   ;


input_declaration_no_semi :
    INPUT
    (
        (
          WIRE
        )
        |
        (
          REG
        )
        |
        (
          INTEGER
        )
        |
        (
          REAL
        )
     )?
    (
      'signed'
    )? 
    (
        range
    )? 
    list_of_variables
    ;

output_declaration_no_semi:
    (OUTPUT (REG)? ('signed')? (range)? IDENTIFIER ASSIGN) => output_declaration_with_assign
    |
    OUTPUT 
    (
        (
          'wire'
        )
        |
        (
          REG
        )
        |
        (
          INTEGER
        )
        |
        (
          REAL
        )
     )?    
    (
      'signed'
    )?
    (
        range
    )? 
    list_of_variables  
    ;

output_declaration_with_assign :
    o1=OUTPUT 
    (
        (
          WIRE
        )
        |
        (
          'reg'
        )
        |
        (
          INTEGER
        )
        |
        (
          REAL
        )
     )?    
    (
      'signed'
    )?
    (
        range
    )?
    (
        list_of_variables   
        COMMA
    )*
    l2=list_of_assignments
    ;

inout_declaration_no_semi :
    INOUT 
     (
        (
          WIRE
        )
        |
        (
          REG
        )
        |
        (
          INTEGER
        )
        |
        (
          REAL
        )
     )?
    (
      'signed'
    )?
    (
        range
    )? 
    list_of_variables
    ;



input_declaration :
      input_declaration_no_semi SEMI
      ;


output_declaration : 
  output_declaration_no_semi SEMI
  ;

inout_declaration :
    inout_declaration_no_semi SEMI
    ;

net_declaration :
    (net_type ('signed')? (range)? IDENTIFIER ASSIGN) => net_declaration_with_assign
    |
    n1=net_type ('signed' )? (rng1=range)? 
    vlist=list_of_net_variables
    (   COMMA 
        l2=list_of_assignments
    )*
    SEMI
    ;

    
reg_declaration_with_assign :    
    rg=REG 
    ('signed')? 
    (range)? 
    l2=list_of_assignments
    SEMI
    ;

integer_declaration_with_assign :    
    rg=INTEGER l2=list_of_assignments
    SEMI
    ;
    
net_declaration_with_assign :    
    n2=net_type (drive_strength)? 
    ('signed')? (expandrange)? 
    (delay)? 
    l2=list_of_assignments
    SEMI
    ;

net_type :
    t=WIRE
    | 
    'tri' 
    |
    'tri1'
    |
    'supply0'
    |
    'wand' 
    |
    'triand' 
    |
    'tri0'
    |
    'supply1' 
    |
    'wor' 
    |
    'trior' 
    |
    KW_TRIREG 
    ;

expandrange :
    'scalared' range
    |
    'vectored' range 
    |
    r1=range
    ;

reg_declaration :
    (REG ('signed')? (range)? IDENTIFIER ASSIGN) => reg_declaration_with_assign
    |
    rg=REG ('signed' )? (range)? 
    l1=list_of_register_variables
        (   COMMA
        l2=list_of_assignments
    )?
    SEMI
    ;

time_declaration :
    'time' list_of_register_variables SEMI
    ;

integer_declaration :
    (INTEGER IDENTIFIER ASSIGN) => integer_declaration_with_assign
    |
    INTEGER l1=list_of_register_variables SEMI
    ;


real_declaration :
    (REAL | REALTIME ) 
    vlist=list_of_net_variables SEMI
    ;

event_declaration :
    'event' name_of_event ( COMMA name_of_event )* SEMI
    ;

continuous_assign :              
    KW_ASSIGN (drive_strength)? (delay )? (l1=list_of_assignments) SEMI
    ;

parameter_override :
    'defparam' l1=list_of_param_assignments SEMI
        ;

list_of_variables :
    id1=name_of_variable
    ( COMMA 
        id2=name_of_variable
    )*
    |
    def1=DEFINE
    ( COMMA def2=DEFINE
    )*
    ;

list_of_assigned_variables :
    (v1=name_of_variable
    ( 
        (ASSIGN e1=expression) 
    )?
    )
    ( COMMA
        ( 
            v2=name_of_variable
            ( 
                ASSIGN e2=expression 
            )?
        ) 
    )*
    ;

list_of_register_variables :
    r1=register_variable
    ( COMMA 
        r2=register_variable 
    )*
    |
    def1=DEFINE
    ( COMMA def2=DEFINE
    )*
    ;

register_variable :
    name_of_register 
    |
    name_of_memory LBRACK e1=expression COLON e2=expression RBRACK
    ( 
    LBRACK e3=expression COLON e4=expression RBRACK
    )*
    ;


list_of_net_variables :
    r1=net_variable
    ( COMMA 
        r2=net_variable 
    )*
    |
    def1=DEFINE
    ( 
    COMMA def2=DEFINE
    )*
    ;
    
net_variable :
    name_of_net
    ( 
    LBRACK e3=expression COLON e4=expression RBRACK
    )*
    ;
    
charge_strength :
    LPAREN 'small'  RPAREN 
    |
    LPAREN 'medium' RPAREN 
    |
    LPAREN 'large'  RPAREN
    ;

drive_strength :
    LPAREN strength0 COMMA strength1 RPAREN 
    |
    LPAREN strength1 COMMA strength0 RPAREN
    ;

strength0 :
    'supply0' 
    |
    'strong0' 
    |
    'pull0' 
    |
    'weak0' 
    |
    'highz0'
    ;

strength1 :
    'supply1' 
    |
    'strong1' 
    |
    'pull1' 
    |
    'weak1' 
    |
    'highz1'
    ;


range :
    (LBRACK expression (PLUS|MINUS)? COLON) => 
    LBRACK lt= expression 
    (
        PLUS 
        |
        MINUS 
    )?
    COLON 
    rt=expression
    RBRACK 
    |
    LBRACK bit=expression RBRACK
    ;

    
list_of_assignments :
    asgn1=assignment
    ( 
        COMMA asgn2=assignment 
    )*
    ;


//----------------------------------------------------------------------------
// Primitive Instances
//----------------------------------------------------------------------------

gate_declaration
  :
  gate_type (drive_strength)? (delay)? gate_instance (COMMA gate_instance)* SEMI
  ;

gate_type
  :
  AND
  | NAND
  | OR
  | NOR
  | XOR
  | XNOR
  | BUF
  | BUFIF0
  | BUFIF1
  | NOT
  | NOTIF0
  | NOTIF1
  | PULLDOWN
  | PULLUP
  | NMOS
  | RMOS
  | PMOS
  | RPMOS
  | CMOS
  | RCMOS
  | TRAN
  | RTRAN
  | TRANIF0
  | RTRANIF0
  | TRANIF1
  | RTRANIF1
  ;

delay
  :
  POUND NUMBER
  | POUND identifier
  | POUND LPAREN mintypmax_expression (COMMA mintypmax_expression (COMMA mintypmax_expression)?)? RPAREN
  ;

gate_instance
  :
  (name_of_gate_instance)? LPAREN terminal (COMMA terminal)* RPAREN
  ;

udp_instance
  :
  (name_of_UDP_instance)? LPAREN terminal (COMMA terminal)* RPAREN
  ;

terminal
  :
  expression
  // | IDENTIFIER
  ;

//----------------------------------------------------------------------------
// Module Instantiations
//----------------------------------------------------------------------------

module_instantiation 
  :
  nm=name_of_module  (parameter_value_assignment)? 
      (inst1=module_instance
      {
        if ( null != $nm.tok && null != $inst1.tok )
        {
          SortHDLUnit du = SortHDLRoot.instance().getDesignUnit($nm.tok.getText());          
          if ( null == du )
          {
            du = new SortHDLUnit(SortHDLBase.ObjType.UNDEFINED_MODULE , $nm.tok.getText(), null, $nm.tok.getLine());
            SortHDLRoot.instance().addDesignUnit( du ); 
          }
          if ( ( null != SortHDLRoot.instance().getLastVlogModule() ) && (null != du ) )
          {
             HDLFile file = SortHDLRoot.instance().getHDLFile();
             SortHDLInstance instobj = new SortHDLInstance( 
                                      SortHDLBase.ObjType.INSTANCE,
                                      $inst1.tok.getText(),
                                      null, // Lib name is null
                                      du.getName(),
                                      SortHDLBase.ObjType.MODULE,
                                      file,
                                      $inst1.tok.getLine()
                                      );
            SortHDLRoot.instance().getLastVlogModule().addInstance( instobj );
          }
        } 
     } )
    (COMMA inst2=module_instance
          {
        if ( null != $nm.tok && null != $inst2.tok )
        {
          SortHDLUnit du = SortHDLRoot.instance().getDesignUnit($nm.tok.getText());          
          if ( null == du )
          {
            du = new SortHDLUnit(SortHDLBase.ObjType.UNDEFINED_MODULE , $nm.tok.getText(), null, $nm.tok.getLine());
            SortHDLRoot.instance().addDesignUnit( du ); 
          }
          if ( ( null != SortHDLRoot.instance().getLastVlogModule() ) && (null != du ) )
          {
             HDLFile file = SortHDLRoot.instance().getHDLFile();
             SortHDLInstance instobj = new SortHDLInstance( 
                                      SortHDLBase.ObjType.INSTANCE,
                                      $inst2.tok.getText(),
                                      null, // Lib name is null
                                      du.getName(),
                                      SortHDLBase.ObjType.MODULE,
                                      file,
                                      $inst2.tok.getLine()
                                      );
            SortHDLRoot.instance().getLastVlogModule().addInstance( instobj );
          }
        } 
     }         
    )* SEMI
  ;

parameter_value_assignment
  :
  POUND LPAREN expression (COMMA expression)*   
  | (DOT IDENTIFIER LPAREN expression RPAREN ( COMMA DOT IDENTIFIER LPAREN expression RPAREN  )* )
   RPAREN
  ;

module_instance returns [Token tok;]
  :
  name_of_instance{ $tok = $name_of_instance.tok; } LPAREN list_of_module_connections RPAREN
  ;

list_of_module_connections
  :
  mod_prt_conn1=module_port_connection (COMMA mod_prt_conn2=module_port_connection)*
  | nmd_prt_conn1=named_port_connection (COMMA nmd_prt_conn2=named_port_connection)*
  ;

module_port_connection
  :
  expression
  ;

// expression below isn't optional according to Palnitkar, but
// several examples generated by Cadence use this syntax.

named_port_connection
  :
  DOT IDENTIFIER LPAREN (expression)? RPAREN
  ;

//----------------------------------------------------------------------------
// Behavioral Statements
//----------------------------------------------------------------------------

initial_statement
  :
  INITIAL statement
  ;

always_statement
  :
  ALWAYS statement
  ;

statement_or_null
  :
  (statement) => statement
  | SEMI
  ;

statement
  :
  (lvalue ASSIGN) => blocking_assignment SEMI
  | (lvalue LE) => non_blocking_assignment SEMI
  | conditional_statement
  | case_statement
  | loop_statement
  | procedural_timing_control_statement
  | wait_statement
  | event_trigger
  | seq_block
  | par_block
  | task_enable
  | system_task_enable
  | disable_statement
  | procedural_continuous_assignment
  ;

assignment
  :
  lvalue ASSIGN expression
  ;

blocking_assignment
  :
  lvalue ASSIGN (delay_or_event_control)? expression
  ;

non_blocking_assignment
  :
  lvalue LE (delay_or_event_control)? expression
  ;

// 'else' clause is inherently ambiguous; ANTLR gets it right,
// so suppress warning.

 conditional_statement :
        IF LPAREN expression RPAREN statement_or_null
        ( : ELSE statement_or_null)?
        ;

case_statement
  :
  case_keyword LPAREN expression RPAREN (case_item)+ ENDCASE
  ;

case_keyword
  :
  CASE
  | CASEZ
  | CASEX
  ;

case_item
  :
  expression (COMMA expression)* COLON statement_or_null
  | DEFAULT (COLON)? statement_or_null
  ;

loop_statement
  :
   FOREVER statement
  | REPEAT LPAREN expression RPAREN statement
  | WHILE LPAREN expression RPAREN statement
  | FOR LPAREN assignment SEMI expression SEMI assignment RPAREN statement
  ;

procedural_timing_control_statement
  :
  delay_or_event_control statement_or_null
  ;

wait_statement
  :
  WAIT LPAREN expression RPAREN statement_or_null
  ;

event_trigger
  :
  TRIGGER name_of_event SEMI
  ;

disable_statement
  :
  DISABLE IDENTIFIER SEMI
  ;

seq_block
  :
  BEGIN (COLON name_of_block (block_declaration)*)? (statement)* END
  ;

par_block
  :
  FORK (COLON name_of_block (block_declaration)*)? (statement)* JOIN
  ;

block_declaration
  :
  parameter_declaration
  | reg_declaration
  | integer_declaration
  | real_declaration
  | time_declaration
  | event_declaration
  ;

task_enable
  :
  name_of_task (LPAREN expression (COMMA (expression)?)* RPAREN)? SEMI
  ;

system_task_enable
  :
  SYSTEM_TASK_NAME (LPAREN expression (COMMA (expression)?)* RPAREN)? SEMI
  ;

procedural_continuous_assignment
  :
  KW_ASSIGN assignment SEMI
  | DEASSIGN lvalue SEMI
  | FORCE assignment SEMI
  | RELEASE lvalue SEMI
  ;

delay_or_event_control
  :
  delay_control
  | event_control
  ;

specify_block
  :
  SPECIFY (specify_item)* ENDSPECIFY
  ;

specify_item
  :
  spec_param_declaration
  | (path_declaration) => path_declaration
  | system_timing_check
  | sdpd
  ;

spec_param_declaration
  :
  SPECPARAM list_of_specparam_assignments SEMI
  ;

list_of_specparam_assignments
  :
  specparam_assignment (COMMA specparam_assignment)*
  ;

specparam_assignment
  :
  identifier ASSIGN expression
  ;

path_declaration
  :
  (simple_path_declaration) => simple_path_declaration SEMI
  | (level_sensitive_path_declaration) => level_sensitive_path_declaration SEMI
  | edge_sensitive_path_declaration SEMI
  ;

simple_path_declaration
  :
  (parallel_path_description) => parallel_path_description ASSIGN path_delay_value
  | full_path_descriptor ASSIGN path_delay_value
  ;

parallel_path_description
  :
  LPAREN specify_terminal_descriptor PPATH specify_terminal_descriptor RPAREN
  ;

full_path_descriptor
  :
  LPAREN list_of_path_terminals FPATH list_of_path_terminals RPAREN
  ;

list_of_path_terminals
  :
  specify_terminal_descriptor (COMMA specify_terminal_descriptor)*
  ;

specify_terminal_descriptor
  :
  (identifier LBRACK expression COLON) => identifier LBRACK expression COLON expression RBRACK
  | (identifier LBRACK) => identifier LBRACK expression RBRACK
  | identifier
  ;

path_delay_value
  :
  (path_delay_expression) => path_delay_expression
  | LPAREN list_of_path_delay_expressions RPAREN
  ;

list_of_path_delay_expressions
  :
  path_delay_expression COMMA path_delay_expression (COMMA path_delay_expression (COMMA path_delay_expression COMMA path_delay_expression COMMA path_delay_expression)?)?
  ;

path_delay_expression
  :
  mintypmax_expression
  ;

system_timing_check
  :
  DOLLAR_SETUP LPAREN timing_check_event COMMA timing_check_event COMMA timing_check_limit (COMMA notify_register)? RPAREN SEMI
  | DOLLAR_HOLD LPAREN timing_check_event COMMA timing_check_event COMMA timing_check_limit (COMMA notify_register)? RPAREN SEMI
  | DOLLAR_PERIOD LPAREN controlled_timing_check_event COMMA timing_check_limit (COMMA notify_register)? RPAREN SEMI
  | DOLLAR_WIDTH LPAREN controlled_timing_check_event COMMA timing_check_limit (COMMA expression COMMA notify_register)? RPAREN SEMI
  | DOLLAR_SKEW LPAREN timing_check_event COMMA timing_check_event COMMA timing_check_limit (COMMA notify_register)? RPAREN SEMI
  | DOLLAR_RECOVERY LPAREN controlled_timing_check_event COMMA timing_check_event COMMA timing_check_limit (COMMA notify_register)? RPAREN SEMI
  | DOLLAR_SETUPHOLD LPAREN timing_check_event COMMA timing_check_event COMMA timing_check_limit COMMA timing_check_limit (COMMA notify_register)? RPAREN SEMI
  ;

timing_check_event
  :
  (timing_check_event_control)? specify_terminal_descriptor (TRIPPLE_AMPERSAND timing_check_condition)?
  ;

controlled_timing_check_event
  :
  timing_check_event_control specify_terminal_descriptor (TRIPPLE_AMPERSAND timing_check_condition)?
  ;

timing_check_event_control
  :
  POSEDGE
  | NEGEDGE
  | edge_control_specifier
  ;

edge_control_specifier
  :
  EDGE LBRACK edge_descriptor (COMMA edge_descriptor)* RBRACK
  ;

edge_descriptor
  :
  ZERO_X
  | ONE_X
  | n=NUMBER { 
		$n.text.equals("01") || $n.text.equals("10")
	}?
  | i=IDENTIFIER { 
		$i.text.equals("x1") || $i.text.equals("x0")
	}?
  ;

timing_check_condition
  :
  scalar_timing_check_condition
  ;

scalar_timing_check_condition
  :
  expression
  ;

timing_check_limit
  :
  expression
  ;

notify_register
  :
  name_of_register
  ;

level_sensitive_path_declaration
  :
  (parallel_level_sensitive_path_description) => parallel_level_sensitive_path_description ASSIGN path_delay_value SEMI
  | full_level_sensitive_path_description ASSIGN path_delay_value SEMI
  ;

parallel_level_sensitive_path_description
  :
  IF LPAREN expression RPAREN LPAREN specify_terminal_descriptor (polarity_operator)? PPATH specify_terminal_descriptor RPAREN
  ;

full_level_sensitive_path_description
  :
  IF LPAREN expression RPAREN LPAREN list_of_path_terminals (polarity_operator)? FPATH list_of_path_terminals RPAREN
  ;

polarity_operator
  :
  PLUS
  | MINUS
  ;

edge_sensitive_path_declaration
  :
  (IF LPAREN expression RPAREN)? LPAREN (edge_identifier)? specify_terminal_descriptor
  (
    PPATH
    | FPATH
  )
  LPAREN
  (
    (list_of_path_terminals) => list_of_path_terminals
    | specify_terminal_descriptor
  )
  (polarity_operator)? COLON data_source_expression RPAREN RPAREN ASSIGN path_delay_value SEMI
  ;

data_source_expression
  :
  expression
  ;

edge_identifier
  :
  POSEDGE
  | NEGEDGE
  ;

sdpd
  :
  IF LPAREN expression RPAREN simple_path_declaration SEMI
  ;

lvalue
  :
  (id1=identifier range)
  | id2=identifier
  | concatenation
  ;

concatenation
  :
  (LCURLY expression LCURLY) => LCURLY e1=expression LCURLY e2=expression (COMMA e3=expression)* RCURLY RCURLY
  | LCURLY e4=expression (COMMA e5=expression)* RCURLY
  ;

mintypmax_expression
  :
  expression (COLON expression COLON expression)?
  ;

exp11
  :
  STRING
  | NUMBER
  | (function_call) => function_call
  | lvalue
  | DEFINE
  ;

exp10
  :
  exp11
  | LPAREN expression RPAREN
  ;

exp9
  :
  exp10
  | unary_operator e9=exp9
  ;

exp8
  :
  e91=exp9 (binary_operator e92=exp9)*
  ;

exp7
  :
  exp8 (QUESTION e71=exp7 COLON e72=exp7)?
  ;

exp0
  :
  exp7
  ;

expression
  :
  exp0
  ;

function_call
  :
  name_of_function LPAREN expression_list RPAREN
  | SYSTEM_TASK_NAME (LPAREN expression_list RPAREN)?
  ;

expression_list
  :
  expression (COMMA expression)*
  ;

unary_operator
  :
  PLUS
  | MINUS
  | LNOT
  | BNOT
  | BAND
  | RNAND
  | BOR
  | RNOR
  | BXOR
  | BXNOR
  ;

binary_operator
  :
  PLUS
  | MINUS
  | STAR
  | DIV
  | MOD
  | EQUAL
  | NOT_EQ
  | EQ_CASE
  | NOT_EQ_CASE
  | LAND
  | LOR
  | LT_
  | LE
  | GT
  | GE
  | BAND
  | BOR
  | BXOR
  | BXNOR
  | SR
  | SL
  ;



name_of_module returns [Token tok;]
  :
  id=local_identifier {$tok = $id.tok; }
  | DEFINE 
  ;

name_of_port
  :
  local_identifier
  ;

name_of_variable
  :
  local_identifier
  ;

name_of_UDP
  :
  local_identifier
  ;

name_of_UDP_instance
  :
  local_identifier
  ;

name_of_event
  :
  local_identifier
  ;

name_of_task
  :
  local_identifier
  ;

real_identifier
  :
  identifier
  ;

name_of_memory
  :
  local_identifier
  ;

net_identifier
  :
  identifier
  ;

name_of_function
  :
  local_identifier
  ;

specparam_identifier
  :
  identifier
  ;

udp_name_of_port
  :
  identifier
  ;

name_of_register
  :
  local_identifier
  ;

name_of_net :local_identifier
    ;

name_of_gate_instance
  :
  local_identifier
  ;

name_of_instance returns [Token tok;]
  :
  local_identifier { $tok = $local_identifier.tok; }
  ;

name_of_block
  :
  local_identifier
  ;

output_terminal_name
  :
  local_identifier
  ;

identifier
  :
  identifier_path
  ;

identifier_path
  :
  id1=local_identifier (DOT id2=local_identifier)*
  ;

local_identifier returns [Token tok;]
  :
  id=IDENTIFIER {$tok = $id;	}
  | eid=ESCAPED_IDENTIFIER {$tok = $id;	}
  ;

delay_control
  :
  POUND NUMBER
  | POUND identifier
  | POUND LPAREN mintypmax_expression RPAREN
  ;

event_control
  :
  AT identifier
  | (AT LPAREN event_expression RPAREN)
  | (AT LPAREN? STAR RPAREN?)
  ;

event_expression
  :
  sub_event_expression (OR sub_event_expression)*
  ;

sub_event_expression
  :
  expression
  | POSEDGE expression
  | NEGEDGE expression
  ;

directive
  :
  define_directive
  | include_directive
  | timescale_directive
  | undef_directive
  | resetall_directive
  | conditional_compilation_directive
  | line_compiler_directive
  ;

define_directive
  :
  TICK_DEFINE IDENTIFIER expression
  ;

include_directive
  :
  TICK_INCLUDE  identifier | STRING 
  ;


timescale_directive
  :
  TICK_TIMESCALE  NUMBER IDENTIFIER DIV NUMBER IDENTIFIER
; 
undef_directive
  :
  TICK_UNDEF identifier
  ;

resetall_directive
  :
  TICK_RESETALL
  ;

conditional_compilation_directive
  :
  (TICK_IFDEF) => ifdef_directive
  | (TICK_IFNDEF) => ifndef_directive
  ;

ifdef_directive
  :
  TICK_IFDEF IDENTIFIER group_of_lines (TICK_ELSIF IDENTIFIER group_of_lines)* (TICK_ELSE group_of_lines)? TICK_ENDIF
  ;

ifndef_directive
  :
  TICK_IFNDEF IDENTIFIER group_of_lines (TICK_ELSIF IDENTIFIER group_of_lines)* (TICK_ELSE group_of_lines)? TICK_ENDIF
  ;

group_of_lines
  :
  (
    statement
    | module_item 
    | description
    | udp_body
    | combinational_entry
    | sequential_entry
    | path_declaration
    | system_timing_check
  )*
  ;

line_compiler_directive
  :
  TICK_LINE NUMBER STRING
  ;

udp_body
  :
  (combinational_body) => combinational_body
  | sequential_body
  ;

combinational_body
  :
  TABLE combinational_entry (combinational_entry)* ENDTABLE
  ;

combinational_entry
  :
  level_input_list COLON output_symbol SEMI
  ;

sequential_entry
  :
  seq_input_list COLON state COLON next_state SEMI
  ;

sequential_body
  :
  (udp_initial_statement)? TABLE sequential_entry (sequential_entry)* ENDTABLE
  ;

seq_input_list
  :
  (level_input_list) => level_input_list
  | edge_input_list
  ;

level_input_list
  :
  (level_symbol)+
  ;

edge_input_list
  :
  (options {greedy=true;}: (level_symbol))* edge_indicator (options {greedy=true;}: (level_symbol))*
  ;

edge_indicator
  :
  LPAREN level_symbol level_symbol RPAREN
  | edge_symbol
  ;

state
  :
  level_symbol
  ;

next_state
  :
  output_symbol
  | MINUS
  ;

output_symbol
  :
  n=NUMBER { $n.text.equals("0") || $n.text.equals("1")}?
  | i=IDENTIFIER { $i.text.equals("x") || $i.text.equals("X")}?
  ;

level_symbol
  :
  QUESTION
  | n=NUMBER { $n.text.equals("0") || $n.text.equals("1")}?
  | i=IDENTIFIER { $i.text.equals("x") || $i.text.equals("X") ||
		  $i.text.equals("b") || $i.text.equals("B") }?
  ;

edge_symbol
  :
  STAR
  | i=IDENTIFIER { $i.text.equals("r") || $i.text.equals("R") ||
		  $i.text.equals("f") || $i.text.equals("F") ||
		  $i.text.equals("p") || $i.text.equals("P") ||
		  $i.text.equals("n") || $i.text.equals("N") }?
  ;

//
//MODULE
//    :
//    mod='module'
//    ;
//    
//ENDMODULE:
//    'endmodule'
//    ;
//
//PARAMETER:
//    'parameter'
//    ;
//
//LOCALPARAM:
//    'localparam'
//    ;
//
//ALWAYS:
//    'always'
//    ;
//
//
//INITIAL:
//    'initial'
//    ;
//        
//INTEGER:
//    'integer'
//    ;
//

//                      
//INPUT:
//    'input'
//    ;
//    
//OUTPUT:
//    'output'
//    ;
//
//INOUT:
//    'inout'
//    ;
//    
//WIRE:
//    'wire'
//   ;    
//    
//FUNCTION:
//    'function'
//    ;    
//
//TASK:
//    'task'
//    ;    


ZERO_X: '0x'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
ONE_X: '1x'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
ALWAYS: 'always'   { isSequentialBlock = true; if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
AND: 'and'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
BEGIN: 'begin'  { isSequentialBlock = true; if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
BUF: 'buf'   { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
BUFIF0 : 'bufif0'   { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
BUFIF1: 'bufif1'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
CASE: 'case'  { isSequentialBlock = true; if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
CASEX: 'casex'  { isSequentialBlock = true; if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
CASEZ: 'casez'  { isSequentialBlock = true; if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
CMOS: 'cmos'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
DEASSIGN: 'deassign'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
DEFAULT: 'default'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
TICK_DEFINE: '`define'  { isCompilerDirective = true; if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
DEFPARAM: 'defparam'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
DISABLE: 'disable'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
EDGE: 'edge'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
TICK_ELSE: '`else'   { isCompilerDirective = true; if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
ELSE: 'else'  { isSequentialBlock = true; if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
TICK_ELSIF: '`elsif'  { isCompilerDirective = true; if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
END: 'end'  { isSequentialBlock = false; if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
ENDCASE: 'endcase'  { isSequentialBlock = false; if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
ENDFUNCTION: 'endfunction'  { isSequentialBlock = false; if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
ENDGENERATE: 'endgenerate'   { isSequentialBlock = false; if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
TICK_ENDIF: '`endif'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; isCompilerDirective = false;};
ENDMODULE: 'endmodule'  { isSequentialBlock = false; if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
ENDPRIMITIVE: 'endprimitive'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
ENDSPECIFY: 'endspecify'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
ENDTABLE: 'endtable'  { isSequentialBlock = false; if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
ENDTASK: 'endtask'  { isSequentialBlock = false; if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
EVENT: 'event'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
FOR: 'for'  { isSequentialBlock = true; if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
FORCE: 'force'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
FOREVER:'forever'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
FORK: 'fork'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
FUNCTION: 'function'  { isSequentialBlock = true; if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
GENERATE: 'generate'   { isSequentialBlock = false; if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
GENVAR: 'genvar'   { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
HIGHZ0: 'highz0'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
HIGHZ1: 'highz1'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
DOLLAR_HOLD: '$hold'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
IF: 'if'  { isSequentialBlock = true;if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
TICK_IFDEF: '`ifdef'  { isCompilerDirective = true; if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
TICK_IFNDEF: '`ifndef'  { isCompilerDirective = true; if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
TICK_INCLUDE: '`include'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
INITIAL: 'initial'  { isSequentialBlock = true; if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
INOUT: 'inout'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
INPUT: 'input'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
INTEGER: 'integer'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
JOIN: 'join'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
LARGE: 'large'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
TICK_LINE: '`line'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
LOCALPARAM: 'localparam'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
MACROMODULE: 'macromodule'   { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
MEDIUM: 'medium'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
MODULE: 'module'   { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
NAND: 'nand'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
NEGEDGE: 'negedge'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
NMOS: 'nmos'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
NOR: 'nor'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
NOT: 'not'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
NOTIF0: 'notif0'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
NOTIF1: 'notif1'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
OR: 'or'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
OUTPUT: 'output'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
PARAMETER: 'parameter'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
DOLLAR_PERIOD: '$period'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
PMOS: 'pmos'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
POSEDGE: 'posedge'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
PRIMITIVE: 'primitive'   { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
PULL0: 'pull0'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
PULL1: 'pull1'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
PULLDOWN: 'pulldown'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
PULLUP: 'pullup'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
RCMOS: 'rcmos'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
REAL: 'real'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
REALTIME: 'realtime'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
DOLLAR_RECOVERY: '$recovery'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
REG: 'reg'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
RELEASE: 'release'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
REPEAT: 'repeat'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
TICK_RESETALL: '`resetall'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; }; 
RMOS: 'rnmos'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
RPMOS: 'rpmos'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
RTRAN: 'rtran'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
RTRANIF0: 'rtranif0'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
RTRANIF1: 'rtranif1'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
SCALARED: 'scalared'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
DOLLAR_SETUP: '$setup'   { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
DOLLAR_SETUPHOLD: '$setuphold'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
DOLLAR_SKEW: '$skew'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
SMALL: 'small'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
SPECIFY: 'specify'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
SPECPARAM: 'specparam'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
STRONG0: 'strong0'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
STRONG1: 'strong1'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
SUPPLY0: 'supply0'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
SUPPLY1: 'supply1'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
TABLE: 'table'  { isSequentialBlock = true; if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
TASK: 'task'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
TICK: '`'   { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
TIME: 'time'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
TICK_TIMESCALE: '`timescale'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
TRAN: 'tran'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
TRANIF0: 'tranif0'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
TRANIF1: 'tranif1'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
TRI: 'tri'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
TRI0: 'tri0'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
TRI1: 'tri1'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
TRIAND: 'triand'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
TRIOR: 'trior'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
KW_TRIREG: 'trireg'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
KW_ASSIGN :'assign' { isAssign = true; if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; }; 
TRIPPLE_AMPERSAND: '&&&'   { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
TICK_UNDEF: '`undef'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
VECTORED: 'vectored'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
WAIT: 'wait'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
WAND: 'wand'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
WEAK0: 'weak0'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
WEAK1: 'weak1'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
WHILE: 'while'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
DOLLAR_WIDTH: '$width'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
WIRE: 'wire'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
WOR: 'wor'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; };
XNOR: 'xnor'   { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; } ;
XOR: 'xor'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; } ;


DEFINE
  : 
  TICK IDENTIFIER
  ;


AT
  :
  '@'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; }
  ;

COLON
  :
  ':'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; }
  ;

COMMA
  :
  ','  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; }
  ;

DOT
  :
  '.'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; }
  ;

ASSIGN
  :
  '='  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; }
  ;

MINUS
  :
  '-'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; }
  ;

LBRACK
  :
  '['  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; }
  ;

RBRACK
  :
  ']'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; }
  ;

LCURLY
  :
  '{'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; }
  ;

RCURLY
  :
  '}'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; }
  ;

LPAREN
  :
  '('  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; }
  ;

RPAREN
  :
  ')'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; }
  ;

POUND
  :
  '#'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; }
  ;

QUESTION
  :
  '?'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; }
  ;

SEMI
  :
  ';'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; isAssign = false;}
  ;

PLUS
  :
  '+'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; }
  ;

LNOT
  :
  '!'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; }
  ;

BNOT
  :
  '~'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; }
  ;

BAND
  :
  '&'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; }
  ;

RNAND
  :
  '~&'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; }
  ;

BOR
  :
  '|'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; }
  ;

RNOR
  :
  '~|'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; }
  ;

BXOR
  :
  '^'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; }
  ;

BXNOR
  :
  '~^'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; }
  | '^~'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; }
  ;

STAR
  :
  '*'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; }
  ;

DIV
  :
  '/'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; }
  ;

MOD
  :
  '%'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; }
  ;

EQUAL
  :
  '=='  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; }
  ;

NOT_EQ
  :
  '!='  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; }
  ;

NOT_EQ_CASE
  :
  '!=='  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; }
  ;

EQ_CASE
  :
  '==='  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; }
  ;

LAND
  :
  '&&'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; }
  ;

LOR
  :
  '||'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; }
  ;

LT_
  :
  '<'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; }
  ;

LE
  :
  '<='  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; }
  ;

GT
  :
  '>'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; }
  ;

GE
  :
  '>='  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; }
  ;

SR
  :
  '>>'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; }
  ;

SL
  :
  '<<'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; }
  ;

TRIGGER
  :
  '->'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; }
  ;

PPATH
  :
  '=>'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; }
  ;

FPATH
  :
  '*>'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; }
  ;

IDENTIFIER
  :
  ((
    'a'..'z'
    | 'A'..'Z'
    | '_'
  )
  (
    'a'..'z'
    | 'A'..'Z'
    | '_'
    | '$'
    | '0'..'9'
  )*)
   { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; }
  ;

ESCAPED_IDENTIFIER
  :
  '\\' (~'\u0020')+
  (
    '\u0020'
    | '\t'
    | '\n'
  )
  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; }
  ;

SYSTEM_TASK_NAME
  :
  '$' IDENTIFIER  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; }
  ;

STRING
  :
  '"'
  (
    ~(
      '"'
      | '\n'
     )
  )*
  '"'  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; }
  ;


fragment
VOCAB
  :
  '\u0003'..'\u007F'
  ;

NUMBER
  :
  (
    (
      (SIZE (WS)*)? BASE (WS)*
      (
        (BIT+)
        | (SIZED_DIGIT)
      )
    )
      => SIZED_NUMBER
    | UNSIZED_NUMBER
  )
  { if (  isSequentialBlock  || isAssign || isCompilerDirective ) $channel = HIDDEN; }
  ;

fragment
SIZED_NUMBER
  :
  (SIZE (WS)*)? BASE (WS)*
  (
    ( (BIT)+)
    |
    (
      SIZED_DIGIT
      (
        SIZED_DIGIT
        | '_'
      )*
    )
  )
  ;

fragment
SIZE
  :
  (DIGIT)+
  ;

fragment
BASE
  :
  '\''
  (
    'd'
    | 'D'
    | 'h'
    | 'H'
    | 'o'
    | 'O'
    | 'b'
    | 'B'
    | 's'
    | 'S'
  )
  ;

fragment
SIZED_DIGIT
  :
  DIGIT
  | BIT
  | HEXDIGIT
  ;

fragment
UNSIZED_NUMBER
  :
  DIGIT
  (
    DIGIT
    | '_'
  )*
  ;

fragment
DIGIT
  :
  ('0'..'9')
  ;

fragment
BIT
  :
  (
    '0'
    | '1'
    | 'x'
    | 'z'
    | 'X'
    | 'Z'
    | '?'
  )
  ;

fragment
HEXDIGIT
  :
  (
    'A'..'F'
    | 'a'..'f'
  )
  ;

fragment
SPACE_OR_TAB
  :
  (
    ' '
    | '\t'
  )+
  ;
  
WS
  :
  SPACE_OR_TAB+ { $channel=HIDDEN; }
  ;

NEW_LINE
  :
  (
    ('\r' '\n')
    | ('\n')
  )+
  { $channel=HIDDEN; }
  ;

ML_COMMENT
  :
  '/*' (options {greedy=false;}: .)* '*/' { $channel=HIDDEN; }
  ;

SL_COMMENT
  :
  '//' (options {greedy=false;}: .)* '\r'? '\n' {  $channel=HIDDEN;  }
  ;
