// TJC implementation of procedure parse_get_word_command
package tjc;
import tcl.lang.*;

public class ParseGetWordCommandCmd extends TJC.CompiledCommand {
    public void cmdProc(
        Interp interp,
        TclObject[] objv)
            throws TclException
    {
        if (!initCmd) {
            inlineCmds = true;
            initCmd(interp);
        }
        CallFrame callFrame = TJC.pushLocalCallFrame(interp, wcmd.ns);
        Var[] compiledLocals =
            TJC.initCompiledLocals(callFrame, 3, compiledLocalsNames);
        try {
        if (objv.length != 3) {
            throw new TclNumArgsException(interp, 1, objv, "script stree");
        }
        setVarScalar(interp, "script", objv[1], compiledLocals, 0);
        setVarScalar(interp, "stree", objv[2], compiledLocals, 1);
        { // Invoke: set range [...]
            TclObject tmp0;
            { // Invoke: lindex $stree 1
                TclObject tmp1 = getVarScalar(interp, "stree", compiledLocals, 1);
                tmp1 = TclList.index(interp, tmp1, 1);
                if ( tmp1 == null ) {
                    interp.resetResult();
                } else {
                    interp.setResult(tmp1);
                }
            } // End Invoke: lindex
            tmp0 = interp.getResult();
            setVarScalar(interp, "range", tmp0, compiledLocals, 2);
        } // End Invoke: set
        { // Invoke: return [...]
            { // Invoke: parse getstring $script $range
                TclObject[] objv2 = TJC.grabObjv(interp, 4);
                TclObject tmp3;
                try {
                    // Arg 0 constant: parse
                    objv2[0] = const0;
                    // Arg 1 constant: getstring
                    objv2[1] = const1;
                    // Arg 2 variable: $script
                    tmp3 = getVarScalar(interp, "script", compiledLocals, 0);
                    tmp3.preserve();
                    objv2[2] = tmp3;
                    // Arg 3 variable: $range
                    tmp3 = getVarScalar(interp, "range", compiledLocals, 2);
                    tmp3.preserve();
                    objv2[3] = tmp3;
                    if ( wcmd_cmdEpoch != wcmd.cmdEpoch ) {
                        updateCmdCache(interp, 0);
                    }
                    TJC.invoke(interp, ((cmdcache1_cmdEpoch == cmdcache1.cmdEpoch)
                        ? cmdcache1.cmd : null), objv2, 0);
                    if ( cmdcache1_cmdEpoch != cmdcache1.cmdEpoch ) {
                        updateCmdCache(interp, 1);
                    }
                } finally {
                    tmp3 = objv2[2];
                    if ( tmp3 != null ) {
                        tmp3.release();
                    }
                    tmp3 = objv2[3];
                    if ( tmp3 != null ) {
                        tmp3.release();
                    }
                    TJC.releaseObjv(interp, objv2, 4);
                }
            } // End Invoke: parse
            if ( true ) { return; }
        } // End Invoke: return
        } catch (TclException te) {
            TJC.checkTclException(interp, te, "parse_get_word_command");
        } finally {
            TJC.popLocalCallFrame(interp, callFrame);
        }
    }

    TclObject const0;
    TclObject const1;

    protected void initConstants(Interp interp) throws TclException {
        const0 = TclString.newInstance("parse");
        const0.preserve(); const0.preserve();
        const1 = TclString.newInstance("getstring");
        const1.preserve(); const1.preserve();
    }

    int wcmd_cmdEpoch = 0;
    WrappedCommand cmdcache1;
    int cmdcache1_cmdEpoch;

    void updateCmdCache(Interp interp, int cacheId) throws TclException {
        String cmdName;
        switch ( cacheId ) {
            case 0: {
                cmdcache1 = TJC.INVALID_COMMAND_CACHE;
                cmdcache1_cmdEpoch = 0;
                wcmd_cmdEpoch = wcmd.cmdEpoch;
                return;
            }
            case 1: {
                cmdName = "parse";
                break;
            }
            default: {
                throw new TclRuntimeError("default: cacheId " + cacheId);
            }
        }
        WrappedCommand lwcmd = TJC.resolveCmd(interp, cmdName);
        int cmdEpoch;
        if ( lwcmd == null ) {
            lwcmd = TJC.INVALID_COMMAND_CACHE;
            cmdEpoch = 0;
        } else {
            cmdEpoch = lwcmd.cmdEpoch;
        }
        switch ( cacheId ) {
            case 1: {
                cmdcache1 = lwcmd;
                cmdcache1_cmdEpoch = cmdEpoch;
                break;
            }
        }
    }

    String[] compiledLocalsNames = {
        "script",
        "stree",
        "range"
    };
} // end class ParseGetWordCommandCmd

